"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const request_1 = require("./request");
const response_1 = require("./response");
const logger_1 = tslib_1.__importDefault(require("../logger"));
const utils_1 = require("../utils");
/**
 * There is no valid reason to have more than one driver instances per page
 */
const pagesSet = new Set();
class PuppeteerDriver {
    constructor({ page }) {
        logger_1.default.debug(`instantiating new driver`);
        if (!page) {
            throw new Error('"page" is not defined');
        }
        if (pagesSet.has(page)) {
            logger_1.default.error(`second driver instantiation on the same page`);
            logger_1.default.error(`(probably you did not stop the mocker before start it again)`);
            throw new Error(`second driver instantiation on the same page`);
        }
        this.page = page;
        this.setRequestInterception(true);
        pagesSet.add(page);
        this.page.on('close', () => {
            this.setRequestInterception(false);
            pagesSet.delete(this.page);
        });
    }
    async setRequestInterception(arg) {
        await this.page.setRequestInterception(arg);
    }
    async onRequest(fn) {
        const handler = async (interceptedRequest) => {
            const timestampWithOrder = utils_1.getTimeStampWithStrictOrder();
            utils_1.loggerTrace(`${interceptedRequest.url()} ← page.on('request') fired`);
            const { request, abort, next, respond } = await request_1.extractPuppeteerRequest(interceptedRequest, timestampWithOrder);
            fn({ request, abort, next, respond });
        };
        // Intercepting all requests and respinding with mocks
        this.page.on('request', handler);
        return async () => {
            await this.setRequestInterception(false);
            pagesSet.delete(this.page);
            this.page.off('request', handler);
        };
    }
    async onResponse(fn) {
        const handler = async (interceptedResponse) => {
            const timestampWithOrder = utils_1.getTimeStampWithStrictOrder();
            const url = interceptedResponse.request().url();
            utils_1.loggerTrace(`${url} → page.on('response') fired`);
            await fn(await response_1.extractPuppeteerResponse(interceptedResponse, timestampWithOrder));
            utils_1.loggerTrace(`${url} → finish`);
        };
        // Intercepting all requests and respinding with mocks
        this.page.on('response', handler);
        return () => this.page.off('response', handler);
    }
    onClose(fn) {
        this.page.on('close', fn);
        return () => this.page.off('close', fn);
    }
}
exports.default = PuppeteerDriver;
