"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.AsyncPendingQueue = exports.getBody = exports.userOptionsToOptions = exports.userInterceptorToInterceptor = exports.blacklist = exports.parseUrl = exports.getFormData = exports.isBodyMatched = exports.getQuery = exports.findInterceptor = exports.isInterceptorMatched = exports.getTimeStampWithStrictOrder = exports.assignResponse = exports.loggerTrace = void 0;
const tslib_1 = require("tslib");
const url_1 = require("url");
const debug_1 = tslib_1.__importDefault(require("debug"));
const lodash_ismatch_1 = tslib_1.__importDefault(require("lodash.ismatch"));
const consts_1 = require("./consts");
const words_hash_1 = require("./words-hash");
const mock_id_1 = tslib_1.__importDefault(require("./mock-id"));
const loggerint = debug_1.default('teremock:utils:interceptor');
const loggerBody = debug_1.default('teremock:utils:bodymatched');
exports.loggerTrace = debug_1.default('teremock:trace');
let orderCounter = 0;
function assignResponse(response1, response2) {
    return {
        ...response1,
        ...response2,
        headers: {
            ...response1.headers,
            ...response2 === null || response2 === void 0 ? void 0 : response2.headers,
        },
    };
}
exports.assignResponse = assignResponse;
const getTimeStampWithStrictOrder = () => {
    return {
        timestamp: Date.now(),
        order: ++orderCounter,
    };
};
exports.getTimeStampWithStrictOrder = getTimeStampWithStrictOrder;
const isInterceptorMatched = (interceptor, request) => {
    const { url, method, resourceType = 'xhr' } = request;
    return Object.keys(interceptor).reduce((acc, key) => {
        const value = interceptor[key];
        // Nothing to compare === wildcard (for given `key`)
        if (typeof value === 'undefined' || value === '*' || ((value && value instanceof Set) && value.has('*'))) {
            return acc;
        }
        switch (key) {
            case 'url':
                return acc && url.includes(value);
            case 'methods':
                return acc && value.has(method.toLowerCase());
            case 'query':
                const query = getQuery(url);
                return (acc &&
                    Object.keys(value).reduce((a, k) => {
                        return a && value[k] === query[k];
                    }, true));
            case 'body':
                return acc && isBodyMatched(value, request);
            case 'resourceTypes':
                return acc && value.has(resourceType.toLowerCase());
            default:
                return acc;
        }
    }, true);
};
exports.isInterceptorMatched = isInterceptorMatched;
const findInterceptor = ({ interceptors, request }) => {
    const matchedMockKey = Object.keys(interceptors).find((key) => {
        loggerint(`checking key`, key);
        const interceptor = interceptors[key];
        return exports.isInterceptorMatched(interceptor, request);
    });
    loggerint(`found an interceptor`, matchedMockKey);
    return typeof matchedMockKey === 'string' ? interceptors[matchedMockKey] : null;
};
exports.findInterceptor = findInterceptor;
// duplicates are not supported
function getQuery(url) {
    const urlObj = new url_1.URL(url);
    urlObj.searchParams.sort();
    const entries = [...urlObj.searchParams];
    return entries.reduce((acc, [key, value]) => {
        acc[key] = value;
        return acc;
    }, {});
}
exports.getQuery = getQuery;
function isBodyMatched(value, request) {
    const { headers } = request;
    if (!headers) {
        return false;
    }
    let formData = {};
    if (headers['content-type'] === 'application/x-www-form-urlencoded') {
        formData = getFormData(request.body);
    }
    else if (headers['content-type'] === 'application/json'
        && typeof request.body === 'string') {
        try {
            formData = JSON.parse(request.body);
        }
        catch (e) {
            return false;
        }
    }
    // If body in UserInterceptor is empty - is considered equal
    if (!Object.keys(value).length) {
        return true;
    }
    const objectsIsMatch = lodash_ismatch_1.default(formData, value);
    if (!objectsIsMatch) {
        loggerBody('Body not matched. Request body: ', formData);
    }
    return objectsIsMatch;
}
exports.isBodyMatched = isBodyMatched;
// problem with decoding string, where spaces are encoded as "+" - it is correct for specification
// x-www-form-urlencoded content type https://www.rfc-editor.org/rfc/rfc1866,
// but JS decodeUri function doesn't decode it. same function exist in third-party libraries:
// https://chromium.googlesource.com/chromium/src.git/+/62.0.3178.1/third_party/google_input_tools/third_party/closure_library/closure/goog/string/string.js?autodive=0%2F%2F%2F%2F#486
function decodeUriWithSpaces(encodedStr) {
    return decodeURIComponent(encodedStr.replace(/\+/g, ' '));
}
function getFormData(body) {
    const result = {};
    if (typeof body === 'string') {
        const params = body.split('&');
        params.forEach(param => {
            const paramsPart = param.split('=');
            result[decodeUriWithSpaces(paramsPart[0])] = decodeUriWithSpaces(paramsPart[1]);
        });
    }
    return result;
}
exports.getFormData = getFormData;
function parseUrl(url) {
    const { origin, pathname } = new url_1.URL(url);
    return {
        query: getQuery(url),
        location: origin + pathname,
    };
}
exports.parseUrl = parseUrl;
function blacklist(source, list) {
    if (typeof source === 'undefined') {
        return source;
    }
    const set = new Set(list.map(l => l.toLowerCase()));
    return Object.keys(source).reduce((acc, key) => {
        if (!set.has(key.toLowerCase())) {
            acc[key] = source[key];
        }
        return acc;
    }, {});
}
exports.blacklist = blacklist;
function userInterceptorToInterceptor(uint, nameArg) {
    const name = uint.name || nameArg || `teremock-add-${words_hash_1.humanize(JSON.stringify(uint), 1)}-`;
    const validName = name.toLowerCase().replace(/[^a-z0-9_-]+/, '');
    if (name.toLowerCase() !== validName) {
        throw new Error(`invalid interceptor name "${name}"! only letters, digits, - and _ are allowed.`);
    }
    const defaultMethods = consts_1.DEFAULT_INTERCEPTOR_CAPTURE.methods;
    const defaultResourceTypes = consts_1.DEFAULT_INTERCEPTOR_CAPTURE.resourceTypes;
    const methods = typeof uint.methods === 'string' ? new Set(uint.methods.toLowerCase().split(',')) : defaultMethods;
    const resourceTypes = typeof uint.resourceTypes === 'string' ? new Set(uint.resourceTypes.toLowerCase().split(',')) : defaultResourceTypes;
    let resp = {};
    if ('response' in uint && typeof uint.response !== 'function') {
        resp = {
            response: {
                url: '<unknown>',
                status: 200,
                ...uint.response,
            },
        };
    }
    return {
        ...consts_1.DEFAULT_INTERCEPTOR_CAPTURE,
        ...uint,
        ...resp,
        methods,
        resourceTypes,
        name: validName,
    };
}
exports.userInterceptorToInterceptor = userInterceptorToInterceptor;
function userOptionsToOptions(defaultOptions, userOptions) {
    const defaultInterceptors = defaultOptions.interceptors;
    const { page, ...restUO } = userOptions;
    let interceptors = defaultInterceptors;
    if (typeof restUO.interceptors !== 'undefined') {
        const userInterceptors = restUO.interceptors;
        const customInterceptors = Object.keys(userInterceptors).reduce((acc, key) => {
            acc[key] = userInterceptorToInterceptor(userInterceptors[key], key);
            return acc;
        }, {});
        // Default interceptors must be in the end, but it should be possible to overwrite them
        interceptors = { ...customInterceptors, ...defaultInterceptors, ...customInterceptors };
    }
    let getMockId = defaultOptions.getMockId;
    if (typeof userOptions.getMockId === 'function') {
        getMockId = (...args) => {
            var _a, _b;
            return (_b = (_a = userOptions === null || userOptions === void 0 ? void 0 : userOptions.getMockId) === null || _a === void 0 ? void 0 : _a.call(userOptions, ...args)) !== null && _b !== void 0 ? _b : mock_id_1.default(...args);
        };
    }
    return {
        ...defaultOptions,
        ...restUO,
        interceptors,
        getMockId,
    };
}
exports.userOptionsToOptions = userOptionsToOptions;
// https://github.com/puppeteer/puppeteer/blob/master/docs/api.md#requestrespondresponse
const getBody = (body) => {
    if (typeof body === 'string' || typeof body === 'undefined') {
        return body;
    }
    return JSON.stringify(body);
};
exports.getBody = getBody;
class AsyncPendingQueue {
    constructor() {
        this.allSettledPromise = Promise.resolve();
        this._errorsCount = 0;
    }
    add(promise) {
        const catchedPromise = promise.catch(() => {
            this._errorsCount += 1;
        });
        this.allSettledPromise = Promise.all([this.allSettledPromise, catchedPromise]).then(() => null);
        return promise;
    }
    async awaitPending() {
        await this.allSettledPromise;
        return this._errorsCount;
    }
}
exports.AsyncPendingQueue = AsyncPendingQueue;
